windowcontrols.overlaid > button > image,
windowcontrols.overlaid > button,
windowcontrols.overlaid > button:hover,
.overlaid button,
button.overlaid {
  box-shadow: initial;
}

windowcontrols.overlaid > button:hover > image,
.overlaid button:hover,
button.overlaid:hover {
  box-shadow: 0 0 0 1px white;
}

.shade {
  background-image: linear-gradient(
    alpha(black, 0.6),
    alpha(black, 0.7),
    black
  );
}
